;;; Mudsync --- Live hackable MUD
;;; Copyright © 2016 Christopher Allan Webber <cwebber@dustycloud.org>
;;;
;;; This file is part of Mudsync.
;;;
;;; Mudsync is free software; you can redistribute it and/or modify it
;;; under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; Mudsync is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Mudsync.  If not, see <http://www.gnu.org/licenses/>.

;;; Game actor
;;; ==========

(define-module (mudsync gameobj)
  #:use-module (mudsync command)
  #:use-module (8sync actors)
  #:use-module (8sync agenda)
  #:use-module (8sync rmeta-slot)
  #:use-module (srfi srfi-1)
  #:use-module (ice-9 control)
  #:use-module (ice-9 format)
  #:use-module (ice-9 match)
  #:use-module (oop goops)
  #:export (<gameobj>

            gameobj-loc
            gameobj-gm

            gameobj-act-init
            gameobj-set-loc!
            gameobj-occupants
            gameobj-self-destruct

            slot-ref-maybe-runcheck
            val-or-run

            dyn-ref))

;;; Gameobj
;;; =======


;;; *all* game components that talk to players should somehow
;;; derive from this class.
;;; And all of them need a GM!

(define-class <gameobj> (<actor>)
  ;; location id
  (loc #:init-value #f
       #:getter gameobj-loc)
  
  ;; Uses a hash table like a set (values ignored)
  (occupants #:init-thunk make-hash-table)

  ;; game master id
  (gm #:init-keyword #:gm
      #:getter gameobj-gm)

  ;; a name to be known by
  (name #:init-keyword #:name
        #:init-value #f)
  (goes-by #:init-keyword #:goes-by
           #:init-value #f)

  (desc #:init-value #f
        #:init-keyword #:desc)

  ;; Commands we can handle
  (commands #:allocation #:each-subclass
            #:init-thunk (build-commands
                          ("take" ((direct-command cmd-take)
                                   (prep-indir-command cmd-take-from
                                                       '("from" "out of"))))
                          ("put" ((prep-indir-command cmd-put-in
                                                      '("in" "inside" "on"))))))

  ;; Commands we can handle by being something's container
  (container-commands #:allocation #:each-subclass
                      #:init-thunk (build-commands))

  ;; Commands we can handle by being contained by something else
  (contained-commands #:allocation #:each-subclass
                      #:init-thunk 
                      (build-commands
                       ("drop" ((direct-command cmd-drop #:obvious? #f)))))

  ;; Most objects are generally visible by default
  (invisible? #:init-value #f
              #:init-keyword #:invisible?)
  ;; TODO: Fold this into a procedure in invisible? similar
  ;;   to take-me? and etc
  (visible-to-player?
   #:init-value (wrap-apply gameobj-visible-to-player?))

  ;; Can be a boolean or a procedure accepting
  ;; (gameobj whos-acting #:key from)
  (take-me? #:init-value #f
            #:init-keyword #:take-me?)
  ;; Can be a boolean or a procedure accepting
  ;; (gameobj whos-acting where)
  (drop-me? #:init-value #t
            #:init-keyword #:drop-me?)
  ;; Can be a boolean or a procedure accepting
  ;; (gameobj whos-acting take-what)
  (take-from-me? #:init-value #f
                 #:init-keyword #:take-from-me?)
  ;; Can be a boolean or a procedure accepting
  ;; (gameobj whos-acting put-what)
  (put-in-me? #:init-value #f
              #:init-keyword #:put-in-me?)

  ;; TODO: Remove this and use actor-alive? instead.
  ;; Set this on self-destruct
  ;; (checked by some "long running" game routines)
  (destructed #:init-value #f)

  (actions #:allocation #:each-subclass
           ;;; Actions supported by all gameobj
           #:init-thunk
           (build-actions
            (init gameobj-act-init)
            ;; Commands for co-occupants
            (get-commands gameobj-get-commands)
            ;; Commands for participants in a room
            (get-container-commands gameobj-get-container-commands)
            ;; Commands for inventory items, etc (occupants of the gameobj commanding)
            (get-contained-commands gameobj-get-contained-commands)

            (get-occupants gameobj-get-occupants)
            (add-occupant! gameobj-add-occupant!)
            (remove-occupant! gameobj-remove-occupant!)
            (get-loc gameobj-act-get-loc)
            (set-loc! gameobj-act-set-loc!)
            (get-name gameobj-get-name)
            (set-name! gameobj-act-set-name!)
            (get-desc gameobj-get-desc)
            (goes-by gameobj-act-goes-by)
            (visible-name gameobj-visible-name)
            (self-destruct gameobj-act-self-destruct)
            (tell gameobj-tell-no-op)
            (assist-replace gameobj-act-assist-replace)
            (ok-to-drop-here? (lambda (gameobj message . _)
                                (<-reply message #t))) ; ok to drop by default
            (ok-to-be-taken-from? gameobj-ok-to-be-taken-from)
            (ok-to-be-put-in? gameobj-ok-to-be-put-in)

            ;; Common commands
            (cmd-take cmd-take)
            (cmd-take-from cmd-take-from)
            (cmd-put-in cmd-put-in)
            (cmd-drop cmd-drop))))


;;; gameobj message handlers
;;; ========================

;; Kind of a useful utility, maybe?
(define (simple-slot-getter slot)
  (lambda (actor message)
    (<-reply message (slot-ref actor slot))))

(define (gameobj-replace-step-occupants actor occupants)
  ;; Snarf all the occupants!
  (display "replacing occupant\n")
  (when occupants
    (for-each
     (lambda (occupant)
       (<-wait occupant 'set-loc!
               #:loc (actor-id actor)))
     occupants)))

(define gameobj-replace-steps*
  (list gameobj-replace-step-occupants))

(define (run-replacement actor replaces replace-steps)
  (when replaces
    (mbody-receive (_ #:key occupants)
        (<-wait replaces 'assist-replace)
      (for-each
       (lambda (replace-step)
         (replace-step actor occupants))
       replace-steps))))

;; @@: This could be kind of a messy way of doing gameobj-act-init
;;   stuff.  If only we had generic methods :(
(define* (gameobj-act-init actor message #:key replace)
  "Your most basic game object init procedure.
Assists in its replacement of occupants if necessary and nothing else."
  (run-replacement actor replace gameobj-replace-steps*))

(define (gameobj-goes-by gameobj)
  "Find the name we go by.  Defaults to #:name if nothing else provided."
  (cond ((slot-ref gameobj 'goes-by) =>
         identity)
        ((slot-ref gameobj 'name) =>
         (lambda (name)
           (list name)))
        (else '())))

(define (gameobj-act-goes-by actor message)
  "Reply to a message requesting what we go by."
  (<-reply message #:goes-by (gameobj-goes-by actor)))

(define (val-or-run val-or-proc)
  "Evaluate if a procedure, or just return otherwise"
  (if (procedure? val-or-proc)
      (val-or-proc)
      val-or-proc))

(define (get-candidate-commands actor rmeta-sym verb)
  (class-rmeta-ref (class-of actor) rmeta-sym verb
                   #:dflt '()))

(define* (gameobj-get-commands actor message #:key verb)
  "Get commands a co-occupant of the room might execute for VERB"
  (define candidate-commands
    (get-candidate-commands actor 'commands verb))
  (<-reply message
           #:commands candidate-commands
           #:goes-by (gameobj-goes-by actor)))

(define* (gameobj-get-container-commands actor message #:key verb)
  "Get commands as the container / room of message's sender"
  (define candidate-commands
    (get-candidate-commands actor 'container-commands verb))
  (<-reply message #:commands candidate-commands))

(define* (gameobj-get-contained-commands actor message #:key verb)
  "Get commands as being contained (eg inventory) of commanding gameobj"
  (define candidate-commands
    (get-candidate-commands actor 'contained-commands verb))
  (<-reply message
           #:commands candidate-commands
           #:goes-by (gameobj-goes-by actor)))

(define* (gameobj-add-occupant! actor message #:key who)
  "Add an actor to our list of present occupants"
  (hash-set! (slot-ref actor 'occupants)
             who #t))

(define* (gameobj-remove-occupant! actor message #:key who)
  "Remove an occupant from the room."
  (hash-remove! (slot-ref actor 'occupants) who))

(define* (gameobj-occupants gameobj #:key exclude)
  (hash-fold
   (lambda (occupant _ prev)
     (define exclude-it?
       (match exclude
         ;; Empty list and #f are non-exclusion
         (() #f)
         (#f #f)
         ;; A list of addresses... since our address object is (annoyingly)
         ;; currently a simple cons cell...
         ((exclude-1 ... exclude-rest)
          (member occupant exclude))
         ;; Must be an individual address!
         (_ (equal? occupant exclude))))
     (if exclude-it?
         prev
         (cons occupant prev)))
   '()
   (slot-ref gameobj 'occupants)))

(define* (gameobj-get-occupants actor message #:key exclude)
  "Get all present occupants of the room."
  (define occupants
    (gameobj-occupants actor #:exclude exclude))
  (<-reply message occupants))

(define (gameobj-act-get-loc actor message)
  (<-reply message (slot-ref actor 'loc)))

(define (gameobj-set-loc! gameobj loc)
  "Set the location of this object."
  (define old-loc (gameobj-loc gameobj))
  (format #t "DEBUG: Location set to ~s for ~s\n"
          loc (actor-id-actor gameobj))

  (when (not (equal? old-loc loc))
    (slot-set! gameobj 'loc loc)
    ;; Change registation of where we currently are
    (if old-loc
        (<-wait old-loc 'remove-occupant! #:who (actor-id gameobj)))
    (if loc
        (<-wait loc 'add-occupant! #:who (actor-id gameobj)))))

;; @@: Should it really be #:id ?  Maybe #:loc-id or #:loc?
(define* (gameobj-act-set-loc! actor message #:key loc)
  "Action routine to set the location."
  (gameobj-set-loc! actor loc))

(define (slot-ref-maybe-runcheck gameobj slot whos-asking . other-args)
  "Do a slot-ref on gameobj, evaluating it including ourselves
and whos-asking, and see if we should just return it or run it."
  (match (slot-ref gameobj slot)
    ((? procedure? slot-val-proc)
     (apply slot-val-proc gameobj whos-asking other-args))
    (anything-else anything-else)))

(define gameobj-get-name (simple-slot-getter 'name))

(define* (gameobj-act-set-name! actor message val)
  (slot-set! actor 'name val))

(define* (gameobj-get-desc actor message #:key whos-looking)
  (define desc-text
    (match (slot-ref actor 'desc)
      ((? procedure? desc-proc)
       (desc-proc actor whos-looking))
      (desc desc)))
  (<-reply message desc-text))

(define (gameobj-visible-to-player? gameobj whos-looking)
  "Check to see whether we're visible to the player or not.
By default, this is whether or not the generally-visible flag is set."
  (not (slot-ref gameobj 'invisible?)))

(define* (gameobj-visible-name actor message #:key whos-looking)
  ;; Are we visible?
  (define we-are-visible
    ((slot-ref actor 'visible-to-player?) actor whos-looking))

  (define name-to-return
    (if we-are-visible
        ;; Return our name
        (match (slot-ref actor 'name)
          ((? procedure? name-proc)
           (name-proc actor whos-looking))
          ((? string? name)
           name)
          (#f #f))
        #f))
  (<-reply message #:text name-to-return))

(define (gameobj-self-destruct gameobj)
  "General gameobj self destruction routine"
  ;; Unregister from being in any particular room
  (gameobj-set-loc! gameobj #f)
  (slot-set! gameobj 'destructed #t)
  ;; Boom!
  (self-destruct gameobj))

(define* (gameobj-act-self-destruct gameobj message #:key why)
  "Action routine for self destruction"
  (gameobj-self-destruct gameobj))

;; Unless an actor has a tell message, we just ignore it
(define gameobj-tell-no-op
  (const 'no-op))

(define (gameobj-replace-data-occupants gameobj)
  "The general purpose list of replacement data"
  (list #:occupants (hash-map->list (lambda (occupant _) occupant)
                                    (slot-ref gameobj 'occupants))))

(define (gameobj-replace-data* gameobj)
  ;; For now, just call gameobj-replace-data-occupants.
  ;; But there may be more in the future!
  (gameobj-replace-data-occupants gameobj))

;; So sad that objects must assist in their replacement ;_;
;; But that's life in a live hacked game!
(define (gameobj-act-assist-replace gameobj message)
  "Vanilla method for assisting in self-replacement for live hacking"
  (apply <-reply message
         (gameobj-replace-data* gameobj)))

(define (gameobj-ok-to-be-taken-from gameobj message whos-acting)
  (<-reply message (slot-ref-maybe-runcheck gameobj 'take-me?
                                            whos-acting #:from #t)))

(define (gameobj-ok-to-be-put-in gameobj message whos-acting where)
  (<-reply message (slot-ref-maybe-runcheck gameobj 'drop-me?
                                            whos-acting where)))


;;; Utilities every gameobj has
;;; ---------------------------

(define (dyn-ref gameobj special-symbol)
  "Dynamically look up a special object from the gm"
  (match special-symbol
    ;; if it's a symbol, look it up dynamically
    ((? symbol? _)
     (mbody-val (<-wait (slot-ref gameobj 'gm) 'lookup-special
                        #:symbol special-symbol)))
    ;; if it's false, return nothing
    (#f #f)
    ;; otherwise it's probably an address, return it as-is
    (_ special-symbol)))



;;; Basic actions
;;; -------------

(define* (cmd-take gameobj message #:key direct-obj)
  (define player (message-from message))
  (define player-name
    (mbody-val (<-wait player 'get-name)))
  (define player-loc
    (mbody-val (<-wait player 'get-loc)))
  (define our-name (slot-ref gameobj 'name))
  (define self-should-take
    (slot-ref-maybe-runcheck gameobj 'take-me? player))
  ;; @@: Is there any reason to allow the room to object in the way
  ;;   that there is for dropping?  It doesn't seem like it.
  ;; TODO: Allow gameobj to customize
  (if self-should-take
      ;; Set the location to whoever's picking us up
      (begin
        (gameobj-set-loc! gameobj player)
        (<- player 'tell
            #:text (format #f "You pick up ~a.\n"
                           our-name))
        (<- player-loc 'tell-room
            #:text (format #f "~a picks up ~a.\n"
                           player-name
                           our-name)
            #:exclude player))
      (<- player 'tell
          #:text (format #f "It doesn't seem like you can take ~a.\n"
                         our-name))))

(define* (cmd-drop gameobj message #:key direct-obj)
  (define player (message-from message))
  (define player-name
    (mbody-val (<-wait player 'get-name)))
  (define player-loc
    (mbody-val (<-wait player 'get-loc)))
  (define our-name (slot-ref gameobj 'name))
  (define should-drop
    (slot-ref-maybe-runcheck gameobj 'drop-me? player))
  (define (room-objection-to-drop)
    (mbody-receive (_ drop-ok? #:key why-not) ; does the room object to dropping?
        (<-wait player-loc 'ok-to-drop-here? player (actor-id gameobj))
      (and (not drop-ok?)
           ;; Either give the specified reason, or give a boilerplate one
           (or why-not
               `("You'd love to drop " ,our-name
                 " but for some reason it doesn't seem like you can"
                 " do that here.")))))
  (cond
   ((not player-loc)
    (<- player 'tell
        #:text `("It doesn't seem like you can drop " ,our-name 
                " here, because you don't seem to be anywhere?!?")))
   ;; TODO: Let ourselves supply a reason why not.
   ((not should-drop)
    (<- player 'tell
        #:text (format #f "It doesn't seem like you can drop ~a.\n"
                       our-name)))
   ((room-objection-to-drop)
    (<- player 'tell
        #:text room-objection-to-drop))
   (else
    (gameobj-set-loc! gameobj player-loc)
    ;; TODO: Allow more flavortext here.
    (<- player 'tell
        #:text (format #f "You drop ~a.\n"
                       our-name))
    (<- player-loc 'tell-room
        #:text (format #f "~a drops ~a.\n"
                       player-name
                       our-name)
        #:exclude player))))

(define* (cmd-take-from gameobj message
                        #:key direct-obj indir-obj preposition)
  (define player (message-from message))
  (define player-name
    (mbody-val (<-wait player 'get-name)))
  (define player-loc
    (mbody-val (<-wait player 'get-loc)))
  (define our-name (slot-ref gameobj 'name))
  ;; We need to check if we even have such a thing
  (define thing-to-take
    (call/ec
     (lambda (return)
       (for-each (lambda (occupant)
                   (mbody-receive (_ #:key goes-by)
                       (<-wait occupant 'goes-by)
                     (when (member direct-obj goes-by)
                       (return occupant))))
                 (gameobj-occupants gameobj))
       ;; nothing found
       #f)))
  (define (should-take-from-me)
    (and thing-to-take
         (slot-ref-maybe-runcheck gameobj 'take-from-me? player thing-to-take)))
  ;; @@: Right now we give the same response to both something not being
  ;;   an occupant and to not being permitted to be removed.  This isn't
  ;;   very rich and maybe not as helpful as it could be.  Right now I'm
  ;;   trying to avoid "leaking" information about if an object isn't there.
  ;;   However maybe by making the different responses as slots which can be
  ;;   set, this wouldn't be a problem. 
 (define default-objection
    "As much as you'd like to take it, it doesn't seem like you can.")
  (define (thing-to-take-objection)
    (mbody-receive (_ taken-ok? #:key why-not) ; does the object object to being removed?
        (<-wait thing-to-take 'ok-to-be-taken-from? player) ; @@ no need to supply from where
      (and (not taken-ok?)
           ;; Either give the specified reason, or give a boilerplate one
           (or why-not default-objection))))
  (cond
   ;; Is it not there, or maybe we won't allow it to be taken?
   ((or (not thing-to-take) (not (should-take-from-me)))
    (<- (message-from message) 'tell
        #:text default-objection))
   ;; the thing we wsant to take itself has objected...
   ((thing-to-take-objection) =>
    (lambda (objection)
      (<- (message-from message) 'tell
          #:text objection)))
   ;; looks like we can take it
   (else
    (let ((thing-to-take-name
           (mbody-val (<-wait thing-to-take 'get-name))))
      ;; Wait to announce to the player just in case settting the location
      ;; errors out or something.  Maybe it's overthinking things, I dunno.
      (<-wait thing-to-take 'set-loc! #:loc player)
      (<- player 'tell
          #:text `("You take " ,thing-to-take-name " from "
                   ,our-name "."))
      (<- player-loc 'tell-room
          #:text `(,player-name " takes " ,thing-to-take-name " from "
                                ,our-name ".")
          #:exclude player)))))

(define* (cmd-put-in gameobj message
                     #:key direct-obj indir-obj preposition)
  (define player (message-from message))
  (define player-name
    (mbody-val (<-wait player 'get-name)))
  (define player-loc
    (mbody-val (<-wait player 'get-loc)))
  (define our-name (slot-ref gameobj 'name))
  ;; We need to check if we even have such a thing
  (define this-thing
    (call/ec
     (lambda (return)
       (for-each (lambda (occupant)
                   (mbody-receive (_ #:key goes-by)
                       (<-wait occupant 'goes-by)
                     (when (member direct-obj goes-by)
                       (return occupant))))
                 (mbody-val (<-wait player 'get-occupants)))
       ;; nothing found
       #f)))
  (define (should-put-in-me)
    (and this-thing
         (slot-ref-maybe-runcheck gameobj 'put-in-me? player this-thing)))
  ;; @@: Right now we give the same response to both something not being
  ;;   an occupant and to not being permitted to be removed.  This isn't
  ;;   very rich and maybe not as helpful as it could be.  Right now I'm
  ;;   trying to avoid "leaking" information about if an object isn't there.
  ;;   However maybe by making the different responses as slots which can be
  ;;   set, this wouldn't be a problem. 
  (define default-objection
    "As much as you'd like to, it doesn't seem like you can put that in there.")
  (define (this-thing-objection)
    (mbody-receive (_ put-in-ok? #:key why-not) ; does the object object to being moved?
        (<-wait this-thing 'ok-to-be-put-in? player (actor-id gameobj))
      (and (not put-in-ok?)
           ;; Either give the specified reason, or give a boilerplate one
           (or why-not default-objection))))
  (cond
   ;; Is it not there, or maybe we won't allow it to be taken?
   ((or (not this-thing) (not (should-put-in-me)))
    (<- (message-from message) 'tell
        #:text default-objection))
   ;; the thing we wsant to take itself has objected...
   ((this-thing-objection) =>
    (lambda (objection)
      (<- (message-from message) 'tell
          #:text objection)))
   ;; looks like we can take it
   (else
    (let ((this-thing-name
           (mbody-val (<-wait this-thing 'get-name))))
      

      ;; Wait to announce to the player just in case settting the location
      ;; errors out or something.  Maybe it's overthinking things, I dunno.
      (<-wait this-thing 'set-loc! #:loc (actor-id gameobj))
      (<- player 'tell
          #:text `("You put " ,this-thing-name " in "
                   ,our-name "."))
      (<- player-loc 'tell-room
          #:text `(,player-name " puts " ,this-thing-name " in "
                                ,our-name ".")
          #:exclude player)))))
